@@ -7,11 +7,15 @@ try:
 except ImportError:
     import md5
 
-import sys
+import re
+import warnings
 
 from django import forms
 from django.core.cache import cache as django_cache
+from django.core.exceptions import PermissionDenied
 from django.conf import settings as django_settings
+from django.contrib.contenttypes.models import ContentType
+from django.contrib.sites.models import Site
 from django.contrib import admin
 from django.core.urlresolvers import reverse
 from django.db import models
@@ -19,66 +23,19 @@ from django.db.models import Q, signals
 from django.forms.models import model_to_dict
 from django.forms.util import ErrorList
 from django.http import Http404, HttpResponseRedirect
+from django.utils.datastructures import SortedDict
 from django.utils.safestring import mark_safe
-from django.utils.translation import ugettext_lazy as _, ugettext
+from django.utils.translation import ugettext_lazy as _
 from django.db.transaction import commit_on_success
 
-import mptt
+from mptt.models import MPTTModel
 
 from feincms import settings, ensure_completely_loaded
-from feincms.admin import editor
-from feincms.admin import item_editor
+from feincms.admin import item_editor, tree_editor
 from feincms.management.checker import check_database_schema
-from feincms.models import Base, create_base_model
-from feincms.utils import get_object, copy_model_instance
-import feincms.admin.filterspecs
-
-
-# ------------------------------------------------------------------------
-class ActiveAwareContentManagerMixin(object):
-    """
-    Implement what's necessary to add some kind of "active" state for content
-    objects. The notion of active is defined by a number of filter rules that
-    must all match (AND) for the object to be active.
-
-    A Manager for a content class using the "datepublisher" extension
-    should either adopt this mixin or implement a similar interface.
-    """
-
-    # A list of filters which are used to determine whether a page is active or not.
-    # Extended for example in the datepublisher extension (date-based publishing and
-    # un-publishing of pages)
-    active_filters = ()
-
-    @classmethod
-    def apply_active_filters(cls, queryset):
-        """
-        Return a queryset reflecting the filters defined.
-        """
-        for filt in cls.active_filters:
-            if callable(filt):
-                queryset = filt(queryset)
-            else:
-                queryset = queryset.filter(filt)
-
-        return queryset
-
-    @classmethod
-    def add_to_active_filters(cls, filter):
-        """
-        Add a new clause to the active filters. A filter may be either
-        a Q object to be applied to the content class or a callable taking
-        a queryset and spitting out a new one.
-        """
-        if not cls.active_filters:
-            cls.active_filters = list()
-        cls.active_filters.append(filter)
-
-    def active(self):
-        """
-        Return only currently active objects.
-        """
-        return self.apply_active_filters(self)
+from feincms.models import create_base_model
+from feincms.module.page import processors
+from feincms.utils.managers import ActiveAwareContentManagerMixin
 
 # ------------------------------------------------------------------------
 def path_to_cache_key(path):
@@ -87,28 +44,30 @@ def path_to_cache_key(path):
 
     # logic below borrowed from http://richwklein.com/2009/08/04/improving-django-cache-part-ii/
     # via acdha's django-sugar
-    if len(path) > 250:
+    if len(path) > 200:
         m = md5()
         m.update(path)
-        path = m.hexdigest() + '-' + path[:200]
-
-    cache_key = 'PAGE-FOR-URL-%d-%s' % ( django_settings.SITE_ID, path )
+        path = m.hexdigest() + '-' + path[:180]
 
+    cache_key = 'FEINCMS:%d:PAGE-FOR-URL:%s' % (django_settings.SITE_ID, path)
     return cache_key
 
 class PageManager(models.Manager, ActiveAwareContentManagerMixin):
+    """
+    The page manager. Only adds new methods, does not modify standard Django
+    manager behavior in any way.
+    """
 
-    # The fields which should be excluded when creating a copy. The mptt fields are
-    # excluded automatically by other mechanisms
-    # ???: Then why are the mptt fields listed here?
+    # The fields which should be excluded when creating a copy.
     exclude_from_copy = ['id', 'tree_id', 'lft', 'rght', 'level', 'redirect_to']
 
     def page_for_path(self, path, raise404=False):
         """
-        Return a page for a path.
+        Return a page for a path. Optionally raises a 404 error if requested.
 
-        Example:
-        Page.objects.page_for_path(request.path)
+        Example::
+
+            Page.objects.page_for_path(request.path)
         """
 
         stripped = path.strip('/')
@@ -121,10 +80,8 @@ class PageManager(models.Manager, ActiveAwareContentManagerMixin):
             raise
 
     def page_for_path_or_404(self, path):
-        """
-        Wrapper for page_for_path which raises a Http404 if no page
-        has been found for the passed path.
-        """
+        warnings.warn('page_for_path_or_404 is deprecated. Use page_for_path instead.',
+            DeprecationWarning, stacklevel=2)
         return self.page_for_path(path, raise404=True)
 
     def best_match_for_path(self, path, raise404=False):
@@ -133,10 +90,10 @@ class PageManager(models.Manager, ActiveAwareContentManagerMixin):
         continues to search by chopping path components off the end.
 
         Tries hard to avoid unnecessary database lookups by generating all
-        possible matching URL prefixes and choosing the longtest match.
+        possible matching URL prefixes and choosing the longest match.
 
         Page.best_match_for_path('/photos/album/2008/09') might return the
-        page with url '/photos/album'.
+        page with url '/photos/album/'.
         """
 
         paths = ['/']
@@ -146,11 +103,10 @@ class PageManager(models.Manager, ActiveAwareContentManagerMixin):
         # We flush the cache entry on page saving, so the cache should always
         # be up to date.
 
-        if settings.FEINCMS_USE_CACHE:
-            ck = path_to_cache_key(path)
-            page = django_cache.get(ck)
-            if page:
-                return page
+        ck = path_to_cache_key(path)
+        page = django_cache.get(ck)
+        if page:
+            return page
 
         if path:
             tokens = path.split('/')
@@ -159,8 +115,7 @@ class PageManager(models.Manager, ActiveAwareContentManagerMixin):
         try:
             page = self.active().filter(_cached_url__in=paths).extra(
                 select={'_url_length': 'LENGTH(_cached_url)'}).order_by('-_url_length')[0]
-            if settings.FEINCMS_USE_CACHE:
-                django_cache.set(ck, page)
+            django_cache.set(ck, page)
             return page
         except IndexError:
             if raise404:
@@ -169,87 +124,104 @@ class PageManager(models.Manager, ActiveAwareContentManagerMixin):
         raise self.model.DoesNotExist
 
     def in_navigation(self):
+        """
+        Returns active pages which have the ``in_navigation`` flag set.
+        """
+
         return self.active().filter(in_navigation=True)
 
     def toplevel_navigation(self):
+        """
+        Returns top-level navigation entries.
+        """
+
         return self.in_navigation().filter(parent__isnull=True)
 
-    def for_request(self, request, raise404=False):
-        page = self.page_for_path(request.path, raise404)
-        page.setup_request(request)
-        return page
+    def for_request(self, request, raise404=False, best_match=False, setup=True):
+        """
+        Return a page for the request
+
+        Does not hit the database more than once for the same request.
+
+        Examples::
+
+            Page.objects.for_request(request, raise404=True, best_match=False)
+
+        Defaults to raising a ``DoesNotExist`` exception if no exact match
+        could be determined.
+        """
+
+        if not hasattr(request, '_feincms_page'):
+            path = request.path_info or request.path
+
+            if best_match:
+                request._feincms_page = self.best_match_for_path(path, raise404=raise404)
+            else:
+                request._feincms_page = self.page_for_path(path, raise404=raise404)
+
+        if setup:
+            request._feincms_page.setup_request(request)
+        return request._feincms_page
 
     def for_request_or_404(self, request):
+        warnings.warn('for_request_or_404 is deprecated. Use for_request instead.',
+            DeprecationWarning, stacklevel=2)
         return self.for_request(request, raise404=True)
 
     def best_match_for_request(self, request, raise404=False):
-        page = self.best_match_for_path(request.path, raise404)
+        warnings.warn('best_match_for_request is deprecated. Use for_request instead.',
+            DeprecationWarning, stacklevel=2)
+        page = self.best_match_for_path(request.path, raise404=raise404)
         page.setup_request(request)
         return page
 
-    def from_request(self, request):
+    def from_request(self, request, best_match=False):
+        warnings.warn('from_request is deprecated. Use for_request instead.',
+            DeprecationWarning, stacklevel=2)
+
         if hasattr(request, '_feincms_page'):
             return request._feincms_page
 
+        if best_match:
+            return self.best_match_for_request(request, raise404=False)
         return self.for_request(request)
 
-    def create_copy(self, page):
-        """
-        Creates an identical copy of a page except that the new one is
-        inactive.
-        """
-
-        new = copy_model_instance(page, exclude=self.exclude_from_copy)
-        new.active = False
-        new.save()
-        new.copy_content_from(page)
-
-        return new
-
-    def replace(self, page, with_page):
-        page.active = False
-        page.save()
-        with_page.active = True
-        with_page.save()
-
-        for child in page.children.all():
-            child.parent = Page.objects.get(pk=with_page.pk)
-            child.save()
-
-        # reload to ensure that the mptt attributes in the DB
-        # and in our objects are equal
-        page = Page.objects.get(pk=page.pk)
-        with_page = Page.objects.get(pk=with_page.pk)
-        with_page.move_to(page, 'right')
-
-        return Page.objects.get(pk=with_page.pk)
-
-PageManager.add_to_active_filters( Q(active=True) )
+PageManager.add_to_active_filters(Q(active=True))
 
 # MARK: -
 # ------------------------------------------------------------------------
 
-try:
-    # MPTT 0.4
-    from mptt.models import MPTTModel
-    mptt_register = False
-    Base = create_base_model(MPTTModel)
-except ImportError:
-    # MPTT 0.3
-    mptt_register = True
+class _LegacyProcessorDescriptor(object):
+    """
+    Request and response processors have been moved into their own module;
+    this descriptor allows accessing them the old way (as attributes of the
+    Page class) but emits a warning. This class will only be available during
+    the FeinCMS 1.5 lifecycle.
+    """
+    def __init__(self, name):
+        self.name = name
 
+    def __get__(self, obj, objtype=None):
+        warnings.warn('Page request and response processors have been moved into '
+            'their own module. Accessing them via the Page class will not be possible '
+            'in FeinCMS 1.6 anymore.',
+            DeprecationWarning, stacklevel=2)
+        return getattr(processors, self.name)
 
-class Page(Base):
+    def __set__(self, obj, val):
+        setattr(processors, self.name, val)
 
-    active = models.BooleanField(_('active'), default=False)
+# ------------------------------------------------------------------------
+
+class Page(create_base_model(MPTTModel)):
+    active = models.BooleanField(_('active'), default=True)
 
     # structure and navigation
-    title = models.CharField(_('title'), max_length=200,
-        help_text=_('This is used for the generated navigation too.'))
+    title = models.CharField(_('title'), max_length=200)
     slug = models.SlugField(_('slug'), max_length=150)
     parent = models.ForeignKey('self', verbose_name=_('Parent'), blank=True, null=True, related_name='children')
     parent.parent_filter = True # Custom list_filter - see admin/filterspecs.py
-    in_navigation = models.BooleanField(_('in navigation'), default=False)
+    in_navigation = models.BooleanField(_('in navigation'), default=True)
     override_url = models.CharField(_('override URL'), max_length=300, blank=True,
         help_text=_('Override the target URL. Be sure to include slashes at the beginning and at the end if it is a local URL. This affects both the navigation and subpages\' URLs.'))
     redirect_to = models.CharField(_('redirect to'), max_length=300, blank=True,
@@ -257,8 +229,8 @@ class Page(Base):
     _cached_url = models.CharField(_('Cached URL'), max_length=300, blank=True,
         editable=False, default='', db_index=True)
 
-    request_processors = []
-    response_processors = []
+    request_processors = SortedDict()
+    response_processors = SortedDict()
     cache_key_components = [ lambda p: django_settings.SITE_ID,
                              lambda p: p._django_content_type.id,
                              lambda p: p.id ]
@@ -302,6 +274,8 @@ class Page(Base):
         This is different than page.get_descendants (from mptt) as it will
         additionally select only child pages that are active.
         """
+        warnings.warn('active_children is deprecated. Use self.children.active() instead.',
+            DeprecationWarning, stacklevel=2)
         return Page.objects.active().filter(parent=self)
 
     def active_children_in_navigation(self):
@@ -311,6 +285,8 @@ class Page(Base):
         menues (only show a disclosure indicator if there actually is something
         to disclose).
         """
+        warnings.warn('active_children_in_navigation is deprecated. Use self.children.in_navigation() instead.',
+            DeprecationWarning, stacklevel=2)
         return self.active_children().filter(in_navigation=True)
 
     def short_title(self):
@@ -330,6 +306,12 @@ class Page(Base):
 
     @commit_on_success
     def save(self, *args, **kwargs):
+        """
+        Overridden save method which updates the ``_cached_url`` attribute of
+        this page and all subpages. Quite expensive when called with a page
+        high up in the tree.
+        """
+
         cached_page_urls = {}
 
         # determine own URL
@@ -344,9 +326,8 @@ class Page(Base):
         super(Page, self).save(*args, **kwargs)
 
         # Okay, we changed the URL -- remove the old stale entry from the cache
-        if settings.FEINCMS_USE_CACHE:
-            ck = path_to_cache_key( self._original_cached_url.strip('/') )
-            django_cache.delete(ck)
+        ck = path_to_cache_key( self._original_cached_url.strip('/') )
+        django_cache.delete(ck)
 
         # If our cached URL changed we need to update all descendants to
         # reflect the changes. Since this is a very expensive operation
@@ -355,6 +336,8 @@ class Page(Base):
         if self._cached_url == self._original_cached_url:
             return
 
+        # TODO: Does not find everything it should when ContentProxy content
+        # inheritance has been customized.
         pages = self.get_descendants().order_by('lft')
 
         for page in pages:
@@ -369,19 +352,30 @@ class Page(Base):
             cached_page_urls[page.id] = page._cached_url
             super(Page, page).save() # do not recurse
 
+    @models.permalink
     def get_absolute_url(self):
-        return self._cached_url
+        """
+        Return the absolute URL of this page.
+        """
 
-    def get_preview_url(self):
-        try:
-            return reverse('feincms_preview', kwargs={ 'page_id': self.id })
-        except:
-            return None
+        url = self._cached_url[1:-1]
+        if url:
+            return ('feincms_handler', (url,), {})
+        return ('feincms_home', (), {})
 
     def get_navigation_url(self):
+        """
+        Return either ``redirect_to`` if it is set, or the URL of this page.
+        """
+
         return self.redirect_to or self._cached_url
 
     def get_siblings_and_self(page):
+        """
+        As the name says.
+        """
+        warnings.warn('get_siblings_and_self is deprecated. You probably want self.parent.children.active() anyway.',
+            DeprecationWarning, stacklevel=2)
         return page.get_siblings(include_self=True)
 
     def cache_key(self):
@@ -416,14 +410,43 @@ class Page(Base):
         processor may peruse and modify the page or the request. It can also return
         a HttpResponse for shortcutting the page rendering and returning that response
         immediately to the client.
-        """
-        request._feincms_page = self
-        request._feincms_extra_context = {}
-        request.extra_path = ""
 
-        for fn in self.request_processors:
+        ``setup_request`` stores responses returned by request processors and returns
+        those on every subsequent call to ``setup_request``. This means that
+        ``setup_request`` can be called repeatedly during the same request-response
+        cycle without harm - request processors are executed exactly once.
+        """
+
+        if hasattr(self, '_setup_request_result'):
+            return self._setup_request_result
+        else:
+            # Marker -- setup_request has been successfully run before
+            self._setup_request_result = None
+
+        if not hasattr(request, '_feincms_extra_context'):
+            request._feincms_extra_context = {}
+
+        request._feincms_extra_context.update({
+            'in_appcontent_subpage': False, # XXX This variable name isn't accurate anymore.
+                                            # We _are_ in a subpage, but it isn't necessarily
+                                            # an appcontent subpage.
+            'extra_path': '/',
+            })
+
+        if request.path != self.get_absolute_url():
+            request._feincms_extra_context.update({
+                'in_appcontent_subpage': True,
+                'extra_path': re.sub('^' + re.escape(self.get_absolute_url()[:-1]), '',
+                    request.path),
+                })
+
+        for fn in reversed(self.request_processors.values()):
             r = fn(self, request)
-            if r: return r
+            if r:
+                self._setup_request_result = r
+                break
+
+        return self._setup_request_result
 
     def finalize_response(self, request, response):
         """
@@ -431,147 +454,89 @@ class Page(Base):
         called to modify the response, eg. for setting cache or expiration headers,
         keeping statistics, etc.
         """
-        for fn in self.response_processors:
+        for fn in self.response_processors.values():
             fn(self, request, response)
 
-        if response.status_code == 200 and not response.has_header('Content-Length'):
-            response['Content-Length'] = len(response.content)
-
-    def require_path_active_request_processor(self, request):
-        """
-        Checks whether any ancestors are actually inaccessible (ie. not
-        inactive or expired) and raise a 404 if so.
-        """
-        if not self.are_ancestors_active():
-            raise Http404()
-
     def get_redirect_to_target(self, request):
         """
         This might be overriden/extended by extension modules.
         """
         return self.redirect_to
 
-    def redirect_request_processor(self, request):
-        target = self.get_redirect_to_target(request)
-        if target:
-            return HttpResponseRedirect(target)
-
-    def frontendediting_request_processor(self, request):
-        if not 'frontend_editing' in request.GET:
-            return
-
-        if request.user.has_module_perms('page'):
-            try:
-                enable_fe = int(request.GET['frontend_editing']) > 0
-            except ValueError:
-                enable_fe = False
-
-            request.session['frontend_editing'] = enable_fe
-
-        # Redirect to cleanup URLs
-        return HttpResponseRedirect(request.path)
-
-    def etag_request_processor(self, request):
-
-        # XXX is this a performance concern? Does it create a new class
-        # every time the processor is called or is this optimized to a static
-        # class??
-        class DummyResponse(dict):
-            """
-            This is a dummy class with enough behaviour of HttpResponse so we
-            can use the condition decorator without too much pain.
-            """
-            def has_header(self, what):
-                return False
-
-        def dummy_response_handler(*args, **kwargs):
-            return DummyResponse()
-
-        def etagger(request, page, *args, **kwargs):
-            etag = page.etag(request)
-            return etag
-
-        def lastmodifier(request, page, *args, **kwargs):
-            lm = page.last_modified()
-            return lm
-
-        # Unavailable in Django 1.0 -- the current implementation of ETag support
-        # requires Django 1.1 unfortunately.
-        from django.views.decorators.http import condition
-
-        # Now wrap the condition decorator around our dummy handler:
-        # the net effect is that we will be getting a DummyResponse from
-        # the handler if processing is to continue and a non-DummyResponse
-        # (should be a "304 not modified") if the etag matches.
-        rsp = condition(etag_func=etagger, last_modified_func=lastmodifier)(dummy_response_handler)(request, self)
-
-        # If dummy then don't do anything, if a real response, return and
-        # thus shortcut the request processing.
-        if not isinstance(rsp, DummyResponse):
-            return rsp
-
-    def etag_response_processor(self, request, response):
+    @classmethod
+    def register_request_processor(cls, fn, key=None):
         """
-        Response processor to set an etag header on outgoing responses.
-        The Page.etag() method must return something valid as etag content
-        whenever you want an etag header generated.
+        Registers the passed callable as request processor. A request processor
+        always receives two arguments, the current page object and the request.
         """
-        etag = self.etag(request)
-        if etag is not None:
-            response['ETag'] = '"' + etag + '"'
+        cls.request_processors[fn if key is None else key] = fn
 
-    @staticmethod
-    def debug_sql_queries_response_processor(verbose=False, file=sys.stderr):
-        if not django_settings.DEBUG:
-            return lambda self, request, response: None
-
-        def processor(self, request, response):
-            from django.db import connection
-
-            print_sql = lambda x: x
-            try:
-                import sqlparse
-                print_sql = lambda x: sqlparse.format(x, reindent=True, keyword_case='upper')
-            except:
-                pass
-
-            if verbose:
-                print >> file, "--------------------------------------------------------------"
-            time = 0.0
-            i = 0
-            for q in connection.queries:
-                i += 1
-                if verbose:
-                    print >> file, "%d : [%s]\n%s\n" % ( i, q['time'], print_sql(q['sql']))
-                time += float(q['time'])
-
-            print >> file, "--------------------------------------------------------------"
-            print >> file, "Total: %d queries, %.3f ms" % (i, time)
-            print >> file, "--------------------------------------------------------------"
-
-        return processor
+    @classmethod
+    def register_response_processor(cls, fn, key=None):
+        """
+        Registers the passed callable as response processor. A response processor
+        always receives three arguments, the current page object, the request
+        and the response.
+        """
+        cls.response_processors[fn if key is None else key] = fn
 
     @classmethod
     def register_request_processors(cls, *processors):
-        cls.request_processors[0:0] = processors
+        """
+        Registers all passed callables as request processors. A request processor
+        always receives two arguments, the current page object and the request.
+        """
+
+        warnings.warn("register_request_processors has been deprecated,"
+            " use register_request_processor instead.",
+            DeprecationWarning, stacklevel=2)
+
+        for processor in processors:
+            cls.register_request_processor(processor)
 
     @classmethod
     def register_response_processors(cls, *processors):
-        cls.response_processors.extend(processors)
+        """
+        Registers all passed callables as response processors. A response processor
+        always receives three arguments, the current page object, the request
+        and the response.
+        """
+
+        warnings.warn("register_response_processors has been deprecated,"
+            " use register_response_processor instead.",
+            DeprecationWarning, stacklevel=2)
+
+        for processor in processors:
+            cls.register_response_processor(processor)
 
     @classmethod
     def register_extension(cls, register_fn):
         register_fn(cls, PageAdmin)
 
+    require_path_active_request_processor = _LegacyProcessorDescriptor(
+        'require_path_active_request_processor')
+    redirect_request_processor = _LegacyProcessorDescriptor(
+        'redirect_request_processor')
+    frontendediting_request_processor = _LegacyProcessorDescriptor(
+        'frontendediting_request_processor')
+    etag_request_processor = _LegacyProcessorDescriptor(
+        'etag_request_processor')
+    etag_response_processor = _LegacyProcessorDescriptor(
+        'etag_response_processor')
+    debug_sql_queries_response_processor = _LegacyProcessorDescriptor(
+        'debug_sql_queries_response_processor')
+
 
 # ------------------------------------------------------------------------
-if mptt_register: # MPTT 0.3 legacy support
-    mptt.register(Page)
-
 # Our default request processors
-Page.register_request_processors(Page.require_path_active_request_processor,
-                                 Page.frontendediting_request_processor,
-                                 Page.redirect_request_processor)
+Page.register_request_processor(processors.require_path_active_request_processor,
+    key='path_active')
+Page.register_request_processor(processors.redirect_request_processor,
+    key='redirect')
+
+if settings.FEINCMS_FRONTEND_EDITING:
+    Page.register_request_processor(processors.frontendediting_request_processor,
+        key='frontend_editing')
 
 signals.post_syncdb.connect(check_database_schema(Page, __name__), weak=False)
 
@@ -579,7 +544,7 @@ signals.post_syncdb.connect(check_database_schema(Page, __name__), weak=False)
 # ------------------------------------------------------------------------
 class PageAdminForm(forms.ModelForm):
     never_copy_fields = ('title', 'slug', 'parent', 'active', 'override_url',
-        'translation_of')
+        'translation_of', '_content_title', '_page_title')
 
     def __init__(self, *args, **kwargs):
         ensure_completely_loaded()
@@ -625,7 +590,7 @@ class PageAdminForm(forms.ModelForm):
                             pass
 
                     kwargs['initial'].update(data)
-                except Page.DoesNotExist:
+                except (AttributeError, Page.DoesNotExist):
                     pass
 
         super(PageAdminForm, self).__init__(*args, **kwargs)
@@ -658,6 +623,9 @@ class PageAdminForm(forms.ModelForm):
             current_id = self.instance.id
             active_pages = active_pages.exclude(id=current_id)
 
+        if hasattr(Site, 'page_set') and 'site' in cleaned_data:
+            active_pages = active_pages.filter(site=cleaned_data['site'])
+
         if not cleaned_data['active']:
             # If the current item is inactive, we do not need to conduct
             # further validation. Note that we only check for the flag, not
@@ -691,11 +659,8 @@ class PageAdminForm(forms.ModelForm):
 
         return cleaned_data
 
-
 # ------------------------------------------------------------------------
-
-
-class PageAdmin(editor.ItemEditor, editor.TreeEditor):
+class PageAdmin(item_editor.ItemEditor, tree_editor.TreeEditor):
     class Media:
         css = {}
         js = []
@@ -704,17 +669,21 @@ class PageAdmin(editor.ItemEditor, editor.TreeEditor):
 
     # the fieldsets config here is used for the add_view, it has no effect
     # for the change_view which is completely customized anyway
-    unknown_fields = ['override_url', 'redirect_to']
+    unknown_fields = ['template_key', 'parent', 'override_url', 'redirect_to']
+    fieldset_insertion_index = 2
     fieldsets = [
         (None, {
-            'fields': ['active', 'in_navigation', 'template_key', 'title', 'slug',
-                'parent'],
+            'fields': [
+                ('title', 'slug'),
+                ('active', 'in_navigation'),
+                ],
         }),
-        item_editor.FEINCMS_CONTENT_FIELDSET,
         (_('Other options'), {
             'classes': ['collapse',],
             'fields': unknown_fields,
         }),
+        # <-- insertion point, extensions appear here, see insertion_index above
+        item_editor.FEINCMS_CONTENT_FIELDSET,
         ]
     readonly_fields = []
     list_display = ['short_title', 'is_visible_admin', 'in_navigation_toggle', 'template']
@@ -725,10 +694,19 @@ class PageAdmin(editor.ItemEditor, editor.TreeEditor):
     raw_id_fields = ['parent']
     radio_fields = {'template_key': admin.HORIZONTAL}
 
+    @classmethod
+    def add_extension_options(cls, *f):
+        if isinstance(f[-1], dict):     # called with a fieldset
+            cls.fieldsets.insert(cls.fieldset_insertion_index, f)
+            f[1]['classes'] = list(f[1].get('classes', []))
+            f[1]['classes'].append('collapse')
+        else:   # assume called with "other" fields
+            cls.fieldsets[1][1]['fields'].extend(f)
+
     def __init__(self, *args, **kwargs):
         ensure_completely_loaded()
 
-        if len(Page._feincms_templates) > 4:
+        if len(Page._feincms_templates) > 4 and 'template_key' in self.radio_fields:
             del(self.radio_fields['template_key'])
 
         super(PageAdmin, self).__init__(*args, **kwargs)
@@ -748,14 +726,21 @@ class PageAdmin(editor.ItemEditor, editor.TreeEditor):
                 if not f.editable:
                     self.readonly_fields.append(f.name)
 
-    in_navigation_toggle = editor.ajax_editable_boolean('in_navigation', _('in navigation'))
+    in_navigation_toggle = tree_editor.ajax_editable_boolean('in_navigation', _('in navigation'))
 
     def _actions_column(self, page):
+        editable = getattr(page, 'feincms_editable', True)
+
+        preview_url = "../../r/%s/%s/" % (
+                ContentType.objects.get_for_model(self.model).id,
+                page.id)
         actions = super(PageAdmin, self)._actions_column(page)
-        actions.insert(0, u'<a href="add/?parent=%s" title="%s"><img src="%simg/admin/icon_addlink.gif" alt="%s"></a>' % (
-            page.pk, _('Add child page'), django_settings.ADMIN_MEDIA_PREFIX ,_('Add child page')))
-        actions.insert(0, u'<a href="%s" title="%s"><img src="%simg/admin/selector-search.gif" alt="%s" /></a>' % (
-            page.get_absolute_url(), _('View on site'), django_settings.ADMIN_MEDIA_PREFIX, _('View on site')))
+        if editable:
+            actions.insert(0, u'<a href="add/?parent=%s" title="%s"><img src="%sicon_addlink.gif" alt="%s"></a>' % (
+                page.pk, _('Add child page'), settings._HACK_ADMIN_MEDIA_IMAGES ,_('Add child page')))
+        actions.insert(0, u'<a href="%s" title="%s"><img src="%sselector-search.gif" alt="%s" /></a>' % (
+            preview_url, _('View on site'), settings._HACK_ADMIN_MEDIA_IMAGES, _('View on site')))
+
         return actions
 
     def add_view(self, request, form_url='', extra_context=None):
@@ -772,12 +757,20 @@ class PageAdmin(editor.ItemEditor, editor.TreeEditor):
             response['Location'] += '?parent=%s' % request.GET['parent']
         if 'translation_of' in request.GET:
             # Copy all contents
+            for content_type in obj._feincms_content_types:
+                if content_type.objects.filter(parent=obj).exists():
+                    # Short-circuit processing -- don't copy any contents if
+                    # newly added object already has some
+                    return response
+
             try:
                 original = self.model._tree_manager.get(pk=request.GET.get('translation_of'))
                 original = original.original_translation
                 obj.copy_content_from(original)
                 obj.save()
-            except self.model.DoesNotExist:
+
+                self.message_user(request, _('The content from the original translation has been copied to the newly created page.'))
+            except (AttributeError, self.model.DoesNotExist):
                 pass
 
         return response
@@ -786,53 +779,12 @@ class PageAdmin(editor.ItemEditor, editor.TreeEditor):
         self._visible_pages = list(self.model.objects.active().values_list('id', flat=True))
 
     def change_view(self, request, object_id, extra_context=None):
-        from django.shortcuts import get_object_or_404
-        if 'create_copy' in request.GET:
-            page = get_object_or_404(Page, pk=object_id)
-            new = Page.objects.create_copy(page)
-            self.message_user(request, ugettext("You may edit the copied page below."))
-            return HttpResponseRedirect('../%s/' % new.pk)
-        elif 'replace' in request.GET:
-            page = get_object_or_404(Page, pk=request.GET.get('replace'))
-            with_page = get_object_or_404(Page, pk=object_id)
-            Page.objects.replace(page, with_page)
-            self.message_user(request, ugettext("You have replaced %s. You may continue editing the now-active page below.") % page)
-            return HttpResponseRedirect('.')
-
-        # Hack around a Django bug: raw_id_fields aren't validated correctly for
-        # ForeignKeys in 1.1: http://code.djangoproject.com/ticket/8746 details
-        # the problem - it was fixed for MultipleChoiceFields but not ModelChoiceField
-        # See http://code.djangoproject.com/ticket/9209
-
-        if hasattr(self, "raw_id_fields"):
-            for k in self.raw_id_fields:
-                if not k in request.POST:
-                    continue
-                if not isinstance(getattr(Page, k).field, models.ForeignKey):
-                    continue
-
-                v = request.POST[k]
-
-                if not v:
-                    del request.POST[k]
-                    continue
-
-                try:
-                    request.POST[k] = int(v)
-                except ValueError:
-                    request.POST[k] = None
-
-        return super(PageAdmin, self).change_view(request, object_id, extra_context)
-
-    def render_item_editor(self, request, object, context):
-        if object:
-            try:
-                active = Page.objects.active().exclude(pk=object.pk).get(_cached_url=object._cached_url)
-                context['to_replace'] = active
-            except Page.DoesNotExist:
-                pass
-
-        return super(PageAdmin, self).render_item_editor(request, object, context)
+        try:
+            return super(PageAdmin, self).change_view(request, object_id, extra_context)
+        except PermissionDenied:
+            from django.contrib import messages
+            messages.add_message(request, messages.ERROR, _("You don't have the necessary permissions to edit this object"))
+        return HttpResponseRedirect(reverse('admin:page_page_changelist'))
 
     def is_visible_admin(self, page):
         """
@@ -846,13 +798,13 @@ class PageAdmin(editor.ItemEditor, editor.TreeEditor):
             # parent page's invisibility is inherited
             if page.id in self._visible_pages:
                 self._visible_pages.remove(page.id)
-            return editor.ajax_editable_boolean_cell(page, 'active', override=False, text=_('inherited'))
+            return tree_editor.ajax_editable_boolean_cell(page, 'active', override=False, text=_('inherited'))
 
         if page.active and not page.id in self._visible_pages:
             # is active but should not be shown, so visibility limited by extension: show a "not active"
-            return editor.ajax_editable_boolean_cell(page, 'active', override=False, text=_('extensions'))
+            return tree_editor.ajax_editable_boolean_cell(page, 'active', override=False, text=_('extensions'))
 
-        return editor.ajax_editable_boolean_cell(page, 'active')
+        return tree_editor.ajax_editable_boolean_cell(page, 'active')
     is_visible_admin.allow_tags = True
     is_visible_admin.short_description = _('is active')
     is_visible_admin.editable_boolean_field = 'active'
